#include "fsck.h"

int ro = 0;
int yes = 0;
int preen = 0;
int force = 0;
int all = 0;
const char *dev = NULL;
int h;
size_t size;
void *ptr;

int uncorrected_errors = 0;
int modified = 0;

int main(int argc, char *argv[])
{
	struct stat st;
	int i, r;
	int noopt;

	init_fsck_log();

	noopt = 0;
	for (i = 1; i < argc; i++) {
		if (noopt)
			goto process_dev;
		if (argv[i][0] == '-' && argv[i][1] != '-') {
			int j;
			for (j = 1; argv[i][j]; j++) {
				char c = argv[i][j];
				switch (c) {
					case 'a':
					case 'p':
						preen = 1;
						/* fall through */
					case 'y':
						yes = 1;
						break;
					case 'n':
						ro = 1;
						break;
					case 'r':
						break;
					case 'f':
						force = 1;
						break;
					default:
badsyn:
						fsck_log("Syntax error");
						exit(16);
				}
			}
		} else if (!strcasecmp(argv[i], "--")) noopt = 1;
		else if (argv[i][0] != '-') {
process_dev:
			if (dev)
				goto badsyn;
			dev = argv[i];
		} else
			goto badsyn;
	}
	if (!dev)
		goto badsyn;

	h = open(dev, (ro ? O_RDONLY : O_RDWR));
	if (h == -1)
		fsck_log("Can't open device '%s': %s", dev, strerror(errno)), exit(8);

	if (fstat(h, &st))
		fsck_log("fstat failed: %s", strerror(errno)), exit(8);

	if (S_ISREG(st.st_mode)) {
		size = st.st_size;
		if (size != (uint64_t)st.st_size)
size_truncated_32bit:
			fsck_log("Too big device for 32-bit system"), exit(8);
	} else if (S_ISBLK(st.st_mode)) {
		uint64_t sz64;
		if (ioctl(h, BLKGETSIZE64, &sz64))
			fsck_log("ioctl(BLKGETSIZE64) failed: %s", strerror(errno)), exit(8);
		size = sz64;
		if (size != sz64)
			goto size_truncated_32bit;
	} else {
		fsck_log("The path is neither a block device nor a file");
		exit(8);
	}

	size &= -getpagesize();

	ptr = mmap(NULL, size, PROT_READ | (ro ? 0 : PROT_WRITE), MAP_SHARED, h, 0);
	if (ptr == MAP_FAILED)
		fsck_log("mmap failed: %s", strerror(errno)), exit(8);

	/* Check super block and backup super block */
	if ((r = check_super()))
		goto ret_r;
	/* The super block is valid here */

	if (!force && !le32_to_cpu(sb->dirty) && !le32_to_cpu(sb->error)) {
		fsck_log("the filesystem is clean, no checking needed.");
		goto ret_r;
	}

	if ((r = init_blockmap()))
		goto ret_r;

	/* Check all used inodes */
	if ((r = check_inodes()))
		goto ret_r;
	/* Inodes and their blocks are valid. There are no cross-linked blocks at this point */

	/* Check lost blocks - lost blocks are freed, lost leaf directory nodes have their content reclaimed */
	if ((r = check_lost_blocks()))
		goto ret_r;
	/* Now, there are no lost blocks */

	/* Tree-walk of all the directories from root. Reachable inodes have in-memory inode bit set. Files n_link counts are calculated */
	if ((r = check_connectivity()))
		goto ret_r;
	/* Now, the tree is valid. There may be lost inodes */

	/* Tree unerachable inodes and check i_nlink on reachable ones */
	if ((r = finalize_inodes()))
		goto ret_r;
	/* Now, everything is valid except for free block/inode counts */

	/* Count the number of free blocks/inodes for each group */
	if ((r = finalize_groups()))
		goto ret_r;
	/* Now, everything is valid */

ret_r:
	r |= (uncorrected_errors ? 4 : 0) | (modified ? 2 : 0);

	if (r < 4 && !ro) {
		sb->dirty = cpu_to_le32(0);
		sb->error = cpu_to_le32(0);
	}

	done_blockmap();

	if (!ro) {
		if (msync(ptr, size, MS_SYNC))
			fsck_log("msync failed: %s", strerror(errno)), exit(8);
	}
	if (munmap(ptr, size))
		fsck_log("munmap failed: %s", strerror(errno)), exit(8);
	if (!ro) {
		if (fsync(h))
			fsck_log("fsync failed: %s", strerror(errno)), exit(8);
	}
	if (close(h))
		fsck_log("close failed: %s", strerror(errno)), exit(8);

	done_fsck_log();

	return r;
}
