#include "fsck.h"

int verify_block_clear;

bool verify_block_range(uint64_t b)
{
	if (b & ((1U << sb->log2_page_size) - 1)) {
		return false;
	}
	if (b >> sb->log2_page_size >= le64_to_cpu(sb->n_pages)) {
		return false;
	}
	return true;
}

bool verify_block(uint64_t ino, __le64 *p, bool clone)
{
	uint64_t b;

	b = le64_to_cpu(*p);

	if (verify_block_clear) {
		bmp_op(b >> sb->log2_page_size, B_CLEAR | B_BLOCK | B_MEM);
		bmp_op(b >> sb->log2_page_size, B_CLEAR | B_BLOCK | B_DEV);
		return true;
	}

	if (!verify_block_range(b))
		return false;

	if (!bmp_op(b >> sb->log2_page_size, B_SET | B_BLOCK | B_MEM)) {
		if (clone) {
			if (fsck_query(Q_NORMAL, "inode %"PRIu64" block %"PRIu64" is cross-linked", "clone?", ino, b >> sb->log2_page_size)) {
				uint64_t new_b;
				new_b = alloc_block_or_inode(B_BLOCK);
				if (!new_b)
					return false;
				new_b <<= sb->log2_page_size;
				memcpy(ptr + new_b, ptr + b, 1U << sb->log2_page_size);
				*p = cpu_to_le64(new_b);
				return true;
			}
		}
		return false;
	}

	if (!bmp_op(b >> sb->log2_page_size, B_READ | B_BLOCK | B_DEV)) {
		if (fsck_query(Q_NORMAL, "inode %"PRIu64" block %"PRIu64" is marked as clean", "fix?", ino, b >> sb->log2_page_size)) {
			bmp_op(b >> sb->log2_page_size, B_SET | B_BLOCK | B_DEV);
		}
	}

	return true;
}

void free_block(__le64 *p)
{
	uint64_t b = le64_to_cpu(*p);
	if (!b)
		return;
	bmp_op(b >> sb->log2_page_size, B_CLEAR | B_BLOCK | B_MEM);
	bmp_op(b >> sb->log2_page_size, B_CLEAR | B_BLOCK | B_DEV);
	*p = cpu_to_le64(0);
}
