#include "fsck.h"

int finalize_groups(void)
{
	uint64_t g;

	fsck_log("finalizing groups...");

#pragma omp parallel for
	for (g = 0; g < n_groups; g++) {
		uint32_t free_blocks = 0, free_inodes = 0;
		uint32_t n;
		uint64_t no;
		struct nvfs_group *pg;

		no = g << log2_pages_per_group;
		for (n = 0; n < 1U << log2_pages_per_group; n++) {
			if (!bmp_op(no + n, B_READ | B_BLOCK | B_DEV))
				free_blocks++;
		}

		no = g << log2_inodes_per_group;
		for (n = 0; n < 1U << log2_inodes_per_group; n++) {
			if (!bmp_op(no + n, B_READ | B_INODE | B_DEV))
				free_inodes++;
		}

		pg = p_group(g);

		if (free_blocks != le32_to_cpu(pg->n_free_blocks)) {
			if (fsck_query(Q_PREEN, "the number of free blocks in group %"PRIu64" doesn't match: %u != %u", "fix?", g, (unsigned)free_blocks, (unsigned)le32_to_cpu(pg->n_free_blocks))) {
				pg->n_free_blocks = cpu_to_le32(free_blocks);
			}
		}
		if (free_inodes != le32_to_cpu(pg->n_free_inodes)) {
			if (fsck_query(Q_PREEN, "the number of free inodes in group %"PRIu64" doesn't match: %u != %u", "fix?", g, (unsigned)free_inodes, (unsigned)le32_to_cpu(pg->n_free_inodes))) {
				pg->n_free_inodes = cpu_to_le32(free_inodes);
			}
		}
	}

	return 0;
}
